use std::panic;
use std::io;
use std::io::Result;
use ratatui::{
    buffer::Buffer,
    crossterm::event::{
    	self,
    	Event,
    	KeyCode,
    	KeyEvent,
    	KeyEventKind
    },
    layout::{Alignment, Rect},
    style::Stylize,
    symbols::border,
    text::{Line, Text},
    widgets::{
        block::{Position, Title},
        Block, Paragraph, Widget,
    },
    Frame,
};

mod tui;

#[cfg(test)]
mod test;

#[derive(Debug, Default)]
pub struct App {
	counter: u8,
	exit: bool,
}

fn main() -> Result<()> {
	// Setup panic handling
	let default_panic = panic::take_hook();
	panic::set_hook(Box::new(move |info| {
		let _= tui::restore();
		default_panic(info);
	}));

	let mut terminal = tui::init()?;
	App::default().run(&mut terminal)?;
	tui::restore()?;
	Ok(())
}

impl App {
	// Main app loop until quit
	pub fn run(&mut self, terminal: &mut tui::Tui) -> Result<()> {
		while !self.exit {
			terminal.draw(|frame| self.render_frame(frame))?;
			self.handle_events()?;
		}
		Ok(())
	}
	
	fn render_frame(&self, frame: &mut Frame) {
		frame.render_widget(self, frame.size());
	}
	
	fn handle_events(&mut self) -> io::Result<()> {
		match event::read()? {
			Event::Key(key_event) if key_event.kind == KeyEventKind::Press => {
				self.handle_key_event(key_event)
			}
			_ => {}
		};
		Ok(())
	}

	fn handle_key_event(&mut self, key_event: KeyEvent) {
		match key_event.code {
			KeyCode::Char('q') => self.exit(),
			KeyCode::Left => self.decrement_counter(),
			KeyCode::Right => self.increment_counter(),
			_ => {}
		}
	}

	fn exit(&mut self) {
		self.exit = true;
	}

	fn increment_counter(&mut self) {
		self.counter += 1;
	}

	fn decrement_counter(&mut self) {
		self.counter -= 1;
	}
}

impl Widget for &App {
    fn render(self, area: Rect, buf: &mut Buffer) {
        let title = Title::from(" Counter App Tutorial ".bold());
        let instructions = Title::from(Line::from(vec![
            " Decrement ".into(),
            "<Left>".blue().bold(),
            " Increment ".into(),
            "<Right>".blue().bold(),
            " Quit ".into(),
            "<Q> ".blue().bold(),
        ]));
        let block = Block::bordered()
            .title(title.alignment(Alignment::Center))
            .title(
                instructions
                    .alignment(Alignment::Center)
                    .position(Position::Bottom),
            )
            .border_set(border::THICK);

        let counter_text = Text::from(vec![Line::from(vec![
            "Value: ".into(),
            self.counter.to_string().yellow(),
        ])]);

        Paragraph::new(counter_text)
            .centered()
            .block(block)
            .render(area, buf);
    }
}
